'''
The helper for Ledger output parsing
'''
from typing import List
from cashiersync.model import RegisterRow


class LedgerOutputParser:
    def __init__(self):
        super().__init__()
    
    def get_total_lines(self, output):
        '''
        Extract the total lines from the output, 
        unless there is only one account, in which case use the complete output
        '''
        result = []
        next_line_is_total = False
        total_line = None

        # Special cases
        # if len(output) == 0:
        #     return 0
        
        if len(output) == 1:
            # No income is an array with an empty string ['']
            if output[0] == '':
                total_line = "0"
            else:
                # One-line results don't have totals
                total_line = output[0]
            result.append(total_line)
        else:
            for i, item in enumerate(output):
                # get total
                if next_line_is_total:
                    total_line = output[i]
                    #self.logger.debug(f'total {total_line}')
                    result.append(total_line)
                else:
                    if '------' in output[i]:
                        next_line_is_total = True

        if total_line is None:
            raise ValueError(f'No total fetched in {output}')

        return result

    def remove_blank_values_from_splits(self, split_array: List):
        ''' Removes the empty valuse from a string split array.
        Used mostly when separating by double-space '  ', to clean up.
        '''
        # while '' in split_array:
        #     split_array.remove('')
        res = list(filter(lambda a: a != '', split_array))
        return res

    def trim_all(self, parts: List) -> List:
        ''' Trim all elemnts of a list '''
        return([x.strip() for x in parts])

    def clean_up_register_output(self, report: List[str]) -> List[str]:
        ''' 
        Clean-up the ledger register report.
        The report variable is a list of lines.
        '''
        # eliminate useless lines
        lines = []
        for line in report:
            if line == '':
                pass
            # Check the account line. If empty, skip. This is just the running total.
            elif line[50] == " ":
                # skip
                pass
            else:
                lines.append(line)

        return lines

    def get_rows_from_register(self, ledger_lines: List[str]) -> List[RegisterRow]:
        ''' Parse raw lines from the ledger register output and get RegisterRow. '''
        txs = []
        # previous_tx: RegisterRow = None

        for line in ledger_lines:
            tx = self.get_row_from_register_line(line)

            txs.append(tx)
            # previous_tx = tx

        return txs

    def get_row_from_register_line(self, line: str) -> RegisterRow:
        ''' Parse one register line into a Transaction object '''
        from decimal import Decimal

        has_symbol = line[0:1] != " "

        date_str = line[0:10].strip()
        payee_str = line[11:46].strip()
        account_str = line[46:85].strip()
        amount_str = line[85:107].strip()
            
        tx = RegisterRow()

        # Date
        if date_str == '':
            date_str = None
        tx.date = date_str

        # Payee
        if payee_str == '':
            payee_str = None
        tx.payee = payee_str

        # Symbol
        if has_symbol:
            parts = payee_str.split()
            symbol = parts[0]
            # We need to remove the exchange from ledger symbols, i.e. EUNY.DE
            if '.' in symbol:
                index = symbol.index('.')
                symbol = symbol[0:index]
            tx.symbol = symbol

        # Type
        account = account_str
        # Get just the first 3 characters.
        account = account[0: 2: 1]
        if account == "In":
            tx.type = "Dividend"
        if account == "Ex":
            tx.type = "Tax"

        # Account
        tx.account = account_str

        # Amount
        # Get from the end.
        #index = len(parts) - 4
        parts = amount_str.split()
        #amount_str = parts[index].replace(',', '')
        amount_str = parts[0].replace(',', '')
        tx.amount = Decimal(amount_str)
        # Currency
        #index = len(parts) - 3
        #tx.currency = parts[index]
        tx.currency = parts[1]

        return tx



'''
Encapsulates subprocess for execution of external console commands
and programs.
'''

class Executor:
    def __init__(self, logger = None):
        super().__init__()
        self.logger = logger

    def run(self, command, cwd: str = None):
        ''' Execute a command in the local terminal '''
        import subprocess

        #self.logger.debug(f'running in {cwd}')

        result = subprocess.run(command, shell=True, encoding="utf-8", capture_output=True,
            cwd=cwd)

        # if self.logger:
        #     self.logger.debug(result)

        if result.returncode != 0:
            output = result.stderr
            raise Exception('Ledger error', output)
        else:
            output = result.stdout
        
        #output = self.split_lines(output)

        return output



'''
Git-related operations
Potentially destructive, therefore removed from the API.
'''

# @app.route('/repo/pull', methods=['POST'])
# def repo_pull():
#     '''
#     Pull the changes in the repository.
#     Expects { repoPath: ... } parameter in JSON.
#     '''
#     from cashiersync.executor import Executor

#     json = request.get_json()
#     repo_path = json['repoPath']

#     # Execute git pull and return all the console results.
#     try:
#         executor = Executor(app.logger)
#         output = executor.run("git pull", repo_path)
#     except Exception as e:
#         output = f'Error: {str(e)}'

#     return output

# @app.route('/repo/commit', methods=['POST'])
# def repo_commit():
#     '''
#     Pull the changes in the repository.
#     Expects { repoPath: ... } parameter in JSON.
#     '''
#     from cashiersync.executor import Executor

#     json = request.get_json()
#     repo_path = json['repoPath']
#     message = json['commitMessage']

#     # Execute git pull and return all the console results.
#     try:
#         executor = Executor(app.logger)

#         # Add all the changes.
#         command = f"git add ."
#         output = executor.run(command, repo_path)

#         # Commit
#         command = f'git commit -m "{message}"'
#         output += executor.run(command, repo_path)
#     except Exception as e:
#         output = f'Error: {str(e)}'

#     return output

# @app.route('/repo/push', methods=['POST'])
# def repo_push():
#     '''
#     Push the changes in the repository.
#     Expects { repoPath: ... } parameter in JSON.
#     '''
#     from cashiersync.executor import Executor

#     json = request.get_json()
#     repo_path = json['repoPath']

#     # Execute git pull and return all the console results.
#     try:
#         executor = Executor(app.logger)
#         output = executor.run("git push", repo_path)
#     except Exception as e:
#         output = f'Error: {str(e)}'

#     return output

# @app.route('/append', methods=['POST'])
# def append():
#     ''' Append the text to the given (journal) file '''
#     from cashiersync.executor import Executor
#     import os.path

#     json = request.get_json()
#     file_path = json['filePath']
#     content = json['content']

#     # confirm that the file exists
#     if not os.path.isfile(file_path):
#         raise FileNotFoundError()
    
#     # save
#     with open(file_path, 'a') as destination:
#         destination.write(content)

#     return "Content written"

# @app.route('/repo/status')
# def repo_status():
#     from cashiersync.executor import Executor
    
#     repo_path = request.args.get('repoPath')
#     try:
#         executor = Executor(app.logger)
#         output = executor.run("git status", repo_path)
#     except Exception as e:
#         output = f'Error: {str(e)}'

#     return output



'''
Store configuration in YAML format,
in user's configuration directory.
'''


class Configuration:
    def __init__(self):
        self.config_file_name = "config.yaml"
        self.app_name = "cashiersync"

        self.create_config_file()
        self.config = self.read_config()

    def read_config(self):
        ''' Read configuration file '''
        import yaml

        path = self.get_config_path()
        content = ''
        with open(path, 'r') as stream:
            try:
                content = yaml.safe_load(stream)
                # print(content)
            except yaml.YAMLError as exc:
                print(exc)

        return content

    def create_config_file(self):
        ''' create the config file if it does not exist '''
        import os
        # Create the config folder
        dir = self.get_config_dir()
        if not os.path.exists(dir) and not os.path.isdir(dir):
            os.makedirs(dir)
        # Create file
        path = self.get_config_path()
        if os.path.exists(path):
            return

        with open(path, "w") as config_file:
            content = self.get_template()
            config_file.write(content)

    def get_template(self):
        return '''# Configuration
ledger_working_dir: .
        '''

    # @property
    def get_config_dir(self):
        from xdg.BaseDirectory import xdg_config_home
        from os.path import sep

        return xdg_config_home + sep + self.app_name

    # @property
    def get_config_path(self):
        ''' assembles the path to the config file '''
        from os.path import sep

        return self.get_config_dir() + sep + self.config_file_name

    @property
    def ledger_working_dir(self):
        value = self.config["ledger_working_dir"]
        return value



'''
Parse the output for lots and return the array of lots for a symbol
'''

class LotParser:
    def __init__(self, logger):
        super().__init__()

        self.logger = logger
    
    def get_lots(self, symbol):
        from .ledger_exec import LedgerExecutor
        from .ledger_output_parser import LedgerOutputParser

        params = f"b ^Assets and :{symbol}$ --lots"
    
        ledger = LedgerExecutor(self.logger)
        output = ledger.run(params)
        output = ledger.split_lines(output)

        # parser = LedgerOutputParser()
        # total_lines = parser.get_total_lines(output)

        #
        num_lines = len(output)
        last_line = output[num_lines - 1]
        if "Assets" in last_line:
            #position = last_line.find("Assets")
            #last_line = last_line.substring(0, position)
            parts = last_line.split("Assets")
            last_line = parts[0].strip()
            output[num_lines - 1] = last_line

        return output
